// 6 Digit clock using TPIC6B595 shift registers
// Leftmost digit is digit 0
// Data will be fed in to the rightmost register first so digit 0 needs shifting out first. 
// The clock will have 4 buttons - enter / exit set mode, hour set, minute set and 12 / 24 hour set mode.
// To set the time press the set button and the clock will go to the set time function - this is detected in main loop.
// To exit time set function press set button again, this will zero the seconds and write new values to the rtc.

// version 1.1; added 3 second delay at powerup. Removed leading "0" when hours less than 10. Removed tail from number 7's.
// version 1.2; added ability to disable seconds display by using 4th button to toggle it on and off.
// version 1.3; replaced mode added in V1.2 with 12 hour mode. Used decimal points as colons with inverted three 7 segment displays so the DP is at the top.
// version 1.3A; changed the brighness setting lower as display was extremely bright. Allow leading zero in 24 hour mode. Made 12 hour mode default.
// 7-segment digits 0-9
// {Q0, Q1, Q2, Q3, Q4, Q5, Q6, Q7} --> {A, B, C, D, E, F, G, DP} 

#include <SPI.h>
#include <SoftPWM.h>
#include <Wire.h>                        
#include <RTClib.h>
#include "Button.h"  // Button library by Alexander Brevig                      

const int reg_clock = 13; // SRCK
const int reg_latch = 10; // RCLK
const int reg_data = 11; // SER_IN
const int reg_OE = 9; // output enable; Active LOW. Can be used to control brightness with PWM.
const int colon_pin = 8; // used to dim the colon LED's when night mode enabled.

Button button1 = Button(2,BUTTON_PULLUP_INTERNAL);       // Setup button A (using button library)
Button button2 = Button(3,BUTTON_PULLUP_INTERNAL);       // Setup button B (using button library)
Button button3 = Button(4,BUTTON_PULLUP_INTERNAL);       // Setup button C (using button library)
Button button4 = Button(5,BUTTON_PULLUP_INTERNAL);       // Setup button D (using button library)

int brightness = 2; // set display brightness. Values 1-5. Set to level 4 as LED's I chose were eye searingly bright at level 5.
int brt = 0; // value for storing reversed PWM percentage
int set_mode = 0; // value for storing timeset mode true or false
int twelveH_mode = 0; // value for storing 12 hour mode true or false
int PM = 0;

#define DS3231_I2C_ADDR 0x68
#define DS3231_TEMPERATURE_ADDR 0x11

RTC_DS3231 rtc;
//RTC_Millis rtc; // software rtc for testing

byte ssddigits[10] = // array with decimal points off
{
  B11111100,  // 0
  B01100000,  // 1
  B11011010,  // 2
  B11110010,  // 3
  B01100110,  // 4
  B10110110,  // 5
  B10111110,  // 6
  B11100000,  // 7
  B11111110,  // 8
  B11110110,  // 9
};

byte ssddigitsDP[10] = // array with decimal points on
{
  B11111101,  // 0
  B01100001,  // 1
  B11011011,  // 2
  B11110011,  // 3
  B01100111,  // 4
  B10110111,  // 5
  B10111111,  // 6
  B11100001,  // 7
  B11111111,  // 8
  B11110111,  // 9
};

byte tempC[2] = // C and degrees symbol
{
  B10011100,
  B11000110,
}; 

void setup() // runs once at powerup
{
  
  pinMode (reg_clock, OUTPUT);
  pinMode (reg_latch, OUTPUT);
  pinMode (reg_data, OUTPUT);
  pinMode (reg_OE,OUTPUT);
  pinMode (colon_pin, OUTPUT);
  digitalWrite(reg_OE,LOW); // enable output of shift register chain; an external pullup resistor on the OE pin prevents garbage being displayed at power on.
  digitalWrite(colon_pin, HIGH); // set colons to bright initially
  setBrt();
  SoftPWMBegin();
  SoftPWMSet(9, 0); // init software PWM on pin 9. This pulses the OE pin on and off to control brightness.
  SoftPWMSetPercent(9, brt);
  delay(3000); // put 3 second delay here before starting i2C comms. This allows the DS3231 to do power on calibration before first i2c read/write.
  Wire.begin(); // start i2c 
  rtc.begin(); //start RTC Clock
  Serial.begin(9600);
  set_mode = 0; // set timeset mode to off
  twelveH_mode = 1; // set 12 hour mode on by default
             
  //rtc.adjust(DateTime(F(__DATE__), F(__TIME__))); // comment out once clock is initialy set
    
}

void loop() // main program loop
{
   if (set_mode == 1)

   {
     setTime();
   }

   else if (twelveH_mode == 1 && set_mode == 0)

      {
        twelve_hourMode();
      }

      else

        {
          clockDisplay();
        }
   
   if (button1.uniquePress())
   {
    SPI.beginTransaction(SPISettings(8000000, LSBFIRST, SPI_MODE0)); // display "SEL" for 2 seconds
    digitalWrite(reg_latch,LOW);
    SPI.transfer(B10110110); // S
    SPI.transfer(B10011110); // E
    SPI.transfer(B00011100); // L
    SPI.transfer(B00000000); // blank digit 3
    SPI.transfer(B00000000); // blank seconds display 10s
    SPI.transfer(B00000000); // blank seconds display 1s 
    digitalWrite(reg_latch,HIGH);
    SPI.endTransaction();
    delay(2000);
    set_mode = 1; // enter timeset mode
   }

   if (button4.uniquePress())
   {
    twelveH_mode = !twelveH_mode; // toggle 12 hour mode
   }

  setBrt();
  SoftPWMSetPercent(9, brt); 
}

/// functions ///

void clockDisplay() // gets the time from the rtc and sends data to the shift registers. 24 hour mode.
{

  DateTime now = rtc.now();

  int hours = now.hour();  
	int minutes = now.minute();
  int seconds = now.second();  

  int h1, h2, m1, m2, s1, s2; // split the numbers into seperate digits using mod math
	h2 = hours % 10;
	h1 = ((hours % 100) - h1) / 10;
	m2 = minutes % 10;
	m1 = ((minutes % 100) - m1) / 10;
  s2 = seconds % 10;
  s1 = ((seconds % 100) - s1) / 10;

  // get digit 0
  int dig0 = (h1); // 10h
  // get digit 1
  int dig1 = (h2); // 1h
  // get digit 2
  int dig2 = (m1); // 10m
  // get digit 3
  int dig3 = (m2); // 1m
  // get digit 4
  int dig4 = (s1); // 10s
  // get digit 5
  int dig5 = (s2); // 1s

  SPI.beginTransaction(SPISettings(8000000, LSBFIRST, SPI_MODE0)); // send data to shift registers
  digitalWrite(reg_latch,LOW);

    if ((now.hour() <=9))

      {
      // SPI.transfer(B00000000); // blank leading digit if hours less than 10
       SPI.transfer(ssddigits[dig0]); // show leading zero if hours less then 10
       SPI.transfer(ssddigitsDP[dig1]);
       SPI.transfer(ssddigitsDP[dig2]);
       SPI.transfer(ssddigitsDP[dig3]);
       SPI.transfer(ssddigitsDP[dig4]);
       SPI.transfer(ssddigits[dig5]);
      }

    else

      {
       SPI.transfer(ssddigits[dig0]); 
       SPI.transfer(ssddigitsDP[dig1]);
       SPI.transfer(ssddigitsDP[dig2]);
       SPI.transfer(ssddigitsDP[dig3]);
       SPI.transfer(ssddigitsDP[dig4]);
       SPI.transfer(ssddigits[dig5]);
      }

  digitalWrite(reg_latch,HIGH);

      /*Serial.print(dig0);
      Serial.print(dig1);
      Serial.print(":");
      Serial.print(dig2);
      Serial.print(dig3);
      Serial.print(":");
      Serial.print(dig4);
      Serial.print(dig5);
      Serial.println(" 24 hour mode");*/
    
  SPI.endTransaction();

 if (now.hour() == 23 && now.minute() == 0 && now.second() == 0) // dim display at 23:00
    {
       brightness = 1;
       digitalWrite(colon_pin, LOW); // set colon to dim
    }

    if (now.hour() == 7 && now.minute() == 0 && now.second() == 0) // resume normal brightness at 7:00
    {
       brightness = 2;
       digitalWrite(colon_pin, HIGH); // set colon to bright.
    }

}

void twelve_hourMode() // function for 12 hour mode
{

  DateTime now = rtc.now();
  
    if ((now.hour() >=12)) // sets PM or AM mode depending on 24 hour clock returned by RTC

      PM = 1;

    else

      PM = 0;
  
  int hours = now.hour();  
	int minutes = now.minute();
  int seconds = now.second();  
  
  hours = hours % 12; // convert 24 hours returned by rtc into 12 hours

    if (hours == 0) 

     hours = 12; // at midnight (00 hours) change it to 12 hours. PM indicator will be turned off later in the code below.

  int h1, h2, m1, m2, s1, s2; // split the numbers into seperate digits using mod math
	h2 = hours % 10;
	h1 = ((hours % 100) - h1) / 10;
	m2 = minutes % 10;
	m1 = ((minutes % 100) - m1) / 10;
  s2 = seconds % 10;
  s1 = ((seconds % 100) - s1) / 10;

  // get digit 0
  int dig0 = (h1); // 10h
  // get digit 1
  int dig1 = (h2); // 1h
  // get digit 2
  int dig2 = (m1); // 10m
  // get digit 3
  int dig3 = (m2); // 1m
  // get digit 4
  int dig4 = (s1); // 10s
  // get digit 5
  int dig5 = (s2); // 1s

  SPI.beginTransaction(SPISettings(8000000, LSBFIRST, SPI_MODE0)); // send data to shift registers
  digitalWrite(reg_latch,LOW);
        
  if (hours <=9 && PM == 0) // if its AM and hours are 9 or less.

      {
       SPI.transfer(B00000000); // blank leading digit if hours are less than 9. PM indicator off.
       SPI.transfer(ssddigitsDP[dig1]);
       SPI.transfer(ssddigitsDP[dig2]);
       SPI.transfer(ssddigitsDP[dig3]);
       SPI.transfer(ssddigitsDP[dig4]);
       SPI.transfer(ssddigits[dig5]);
      }

      else if (hours <=9 && PM == 1) // if it's PM and hours are 9 or less

      {
       SPI.transfer(B00000001); // blank leading digit if hours 9 or less and  PM indicator on. 
       SPI.transfer(ssddigitsDP[dig1]);
       SPI.transfer(ssddigitsDP[dig2]);
       SPI.transfer(ssddigitsDP[dig3]);
       SPI.transfer(ssddigitsDP[dig4]);
       SPI.transfer(ssddigits[dig5]);
      }

      else if (hours >=10 && PM == 0) // if it's AM and hours are 10 or more

       {
       SPI.transfer(ssddigits[dig0]); // don't enable PM indicator
       SPI.transfer(ssddigitsDP[dig1]);
       SPI.transfer(ssddigitsDP[dig2]);
       SPI.transfer(ssddigitsDP[dig3]);
       SPI.transfer(ssddigitsDP[dig4]);
       SPI.transfer(ssddigits[dig5]);
       }

      else if (hours >=10 && PM == 1) // if it's PM and hours are 10 or more
      {
       SPI.transfer(ssddigitsDP[dig0]); // enable PM indicator
       SPI.transfer(ssddigitsDP[dig1]);
       SPI.transfer(ssddigitsDP[dig2]);
       SPI.transfer(ssddigitsDP[dig3]);
       SPI.transfer(ssddigitsDP[dig4]);
       SPI.transfer(ssddigits[dig5]);
      }

  digitalWrite(reg_latch,HIGH);
  SPI.endTransaction();

      /*Serial.print(dig0);
      Serial.print(dig1);
      Serial.print(":");
      Serial.print(dig2);
      Serial.print(dig3);
      Serial.print(":");
      Serial.print(dig4);
      Serial.print(dig5);
      Serial.println(" 12 hour mode");*/

 if (now.hour() == 23 && now.minute() == 0 && now.second() == 0) // dim display at 23:00
    {
       brightness = 1;
       digitalWrite(colon_pin, LOW); // set colon to dim
    }

    if (now.hour() == 7 && now.minute() == 0 && now.second() == 0) // resume normal brightness at 7:00
    {
       brightness = 2;
       digitalWrite(colon_pin, HIGH); // set colon to bright.
    }

}

void setTime() // this will run if set time button is pressed

{
    
  DateTime now = rtc.now(); // this gets the current time from the RTC so it updates after button presses.
  int set_min = now.minute();
  int set_hour  = now.hour();
 
  int h1, h2, m1, m2; // split the numbers into seperate digits using mod math
	h2 = set_hour % 10;
	h1 = ((set_hour % 100) - h1) / 10;
	m2 = set_min % 10;
	m1 = ((set_min % 100) - m1) / 10;

  // get digit 0
  int dig0 = (h1); // 10h
  // get digit 1
  int dig1 = (h2); // 1h
  // get digit 2
  int dig2 = (m1); // 10m
  // get digit 3
  int dig3 = (m2); // 1m

  SPI.beginTransaction(SPISettings(8000000, LSBFIRST, SPI_MODE0)); // send data to shift registers
  digitalWrite(reg_latch,LOW);
  SPI.transfer(ssddigits[dig0]); 
  SPI.transfer(ssddigitsDP[dig1]);
  SPI.transfer(ssddigitsDP[dig2]);
  SPI.transfer(ssddigits[dig3]);
  SPI.transfer(B00000000); // blank seconds display 10s
  SPI.transfer(B00000001); // blank seconds display 1s & light last decimal point to indicate set mode
  digitalWrite(reg_latch,HIGH);
  SPI.endTransaction();
     
  // code here will increment values above if hour and minute buttons are pressed also sets button_pressed flag to true

  while (button2.uniquePress())
    {
      set_hour++;
      if (set_hour == 24)
      {
        set_hour = 0;
      }
      delay(10); // 10 millisecond delay to avoid spamming the RTC
      rtc.adjust(DateTime(2023, 10, 23, set_hour, set_min)); // write the time to the rtc, dont zero seconds
    }

  while (button3.uniquePress())
    {
      set_min++;

      if (set_min == 60)
      {
        set_min = 0;
      }
      delay(10); // 10 millisecond delay to avoid spamming the RTC
      rtc.adjust(DateTime(2023, 10, 23, set_hour, set_min)); // write the time to the rtc, don't zero seconds
    }  
   
   if (button1.uniquePress())
    {
      DateTime now = rtc.now(); // this gets the current time from the RTC so it updates after button presses.
      int set_min = now.minute();
      int set_hour  = now.hour();
      rtc.adjust(DateTime(2023, 10, 23, set_hour, set_min,0)); // zero seconds when set button is pressed again
      set_mode = 0; // this happens if timeset button is pressed after adjusting the time
      clockDisplay(); // exit function and return to clock display
    }

}

void setBrt() // this changes the brightness logarithmic reversed scale to a more linear one

{
  if (brightness == 1)
  brt = 99;

  else if (brightness == 2)
  brt = 90;

  else if (brightness == 3)
  brt = 50;

  else if (brightness == 4)
  brt = 30;

  else if (brightness == 5)
  brt = 0;
}

void tempDisplay()
{
  //future function for displaying temperature
}
